from __future__ import annotations

from collections.abc import Callable, Sequence
from typing import Any, Literal, cast

import numpy as np
import pandas as pd
import seaborn as sns
from numpy.typing import ArrayLike
from scipy.sparse import issparse, spmatrix
from sklearn.base import BaseEstimator, is_classifier
from sklearn.inspection import permutation_importance
from sklearn.metrics._scorer import _BaseScorer
from sklearn.pipeline import Pipeline
from sklearn.utils.validation import _num_features

from skore._sklearn._plot.base import BOXPLOT_STYLE, DisplayMixin
from skore._sklearn._plot.inspection.utils import _decorate_matplotlib_axis
from skore._sklearn.feature_names import _get_feature_names
from skore._sklearn.types import Aggregate, DataSource, ReportType
from skore._utils._index import flatten_multi_index


class PermutationImportanceDisplay(DisplayMixin):
    """Display to inspect feature importance via feature permutation.

    Parameters
    ----------
    importances : pd.DataFrame
        The importances computed after permuting the input features. The columns are:

        - `estimator`
        - `data_source`
        - `metric`
        - `split`
        - `feature`
        - `label` or `output` (classification vs. regression)
        - `repetition`
        - `value`

    report_type : {"estimator", "cross-validation"}
        Report type from which the display is created.

    Attributes
    ----------
    facet_ : seaborn FacetGrid
        FacetGrid containing the permutation importance.

    figure_ : matplotlib Figure
        Figure containing the permutation importance.

    ax_ : matplotlib Axes
        Axes with permutation importance.
    """

    _default_boxplot_kwargs: dict[str, Any] = {
        "whis": 1e10,
        **BOXPLOT_STYLE,
    }
    _default_stripplot_kwargs: dict[str, Any] = {
        "palette": "tab10",
        "alpha": 0.5,
        "height": 6,
    }

    def __init__(self, *, importances: pd.DataFrame, report_type: ReportType):
        self.importances = importances
        self.report_type = report_type

    @classmethod
    def _compute_data_for_display(
        cls,
        *,
        data_source: DataSource,
        estimators: Sequence[BaseEstimator],
        names: Sequence[str],
        splits: Sequence[int | float],
        Xs: Sequence[ArrayLike],
        ys: Sequence[ArrayLike],
        at_step: int | str,
        metric: str | Callable | list[str] | tuple[str] | dict[str, Callable] | None,
        n_repeats: int,
        max_samples: float,
        n_jobs: int | None,
        seed: int | None,
        report_type: ReportType,
    ) -> PermutationImportanceDisplay:
        if not isinstance(at_step, str | int):
            raise ValueError(f"at_step must be an integer or a string; got {at_step!r}")

        if isinstance(at_step, str):
            first_estimator = estimators[0]
            if not isinstance(first_estimator, Pipeline):
                raise ValueError(
                    "at_step can only be a string when the estimator is a Pipeline"
                )
            at_step = list(first_estimator.named_steps.keys()).index(at_step)

        all_importances = []
        for estimator, name, split, X, y in zip(
            estimators, names, splits, Xs, ys, strict=True
        ):
            if isinstance(estimator, Pipeline) and at_step != 0:
                if abs(at_step) >= len(estimator.steps):
                    raise ValueError(
                        "at_step must be strictly smaller in magnitude than the "
                        "number of steps in the Pipeline, which is "
                        f"{len(estimator.steps)}; got {at_step}"
                    )
                feature_engineering = estimator[:at_step]
                estimator = estimator[at_step:]
                X_transformed = feature_engineering.transform(X)
            else:
                feature_engineering = None
                X_transformed = X

            feature_names = _get_feature_names(
                estimator,
                n_features=_num_features(X_transformed),
                X=X_transformed,
                transformer=feature_engineering,
            )

            if issparse(X_transformed):
                X_transformed = cast(spmatrix, X_transformed)
                X_transformed = X_transformed.todense()

            scores = permutation_importance(
                estimator=estimator,
                X=X_transformed,
                y=y,
                scoring=metric,
                n_repeats=n_repeats,
                max_samples=max_samples,
                n_jobs=n_jobs,
                random_state=seed,
            )

            if "importances" in scores:
                # single metric case -> switch to multi-metric case by wrapping in a
                # dict with the name of the metric
                metric_obj = cast(str | Callable | _BaseScorer | None, metric)
                if metric_obj is None:
                    metric_name = "accuracy" if is_classifier(estimator) else "r2"
                elif isinstance(metric_obj, str):
                    metric_name = metric_obj
                elif isinstance(metric_obj, _BaseScorer):
                    metric_name = metric_obj._score_func.__name__.replace("_", " ")
                else:
                    metric_name = metric_obj.__name__.replace("_", " ")
                scores = {metric_name: scores}

            df_importances = []
            for metric_name, metric_values in scores.items():
                metric_importances = np.atleast_3d(metric_values["importances"])

                df_metric_importances = []
                # we loop across the labels (for classification) or the outputs
                # (for regression)
                for target_index, target_importances in enumerate(
                    np.moveaxis(metric_importances, -1, 0)
                ):
                    df = pd.DataFrame(
                        target_importances,
                        index=feature_names,
                        columns=range(1, n_repeats + 1),
                    ).melt(var_name="repetition")

                    if metric_importances.shape[-1] == 1:
                        df["label"], df["output"] = np.nan, np.nan
                    else:
                        if is_classifier(estimator):
                            df["label"] = estimator.classes_[target_index]
                            df["output"] = np.nan
                        else:
                            df["output"], df["label"] = target_index, np.nan

                    df["metric"] = metric_name
                    df["feature"] = np.tile(feature_names, n_repeats)
                    df_metric_importances.append(df)

                df_metric_importances = pd.concat(df_metric_importances, axis="index")
                df_importances.append(df_metric_importances)

            df_importances = pd.concat(df_importances, axis="index")
            df_importances["data_source"] = data_source
            df_importances["estimator"] = name
            df_importances["split"] = split
            all_importances.append(df_importances)

        ordered_columns = [
            "estimator",
            "data_source",
            "metric",
            "split",
            "feature",
            "label",
            "output",
            "repetition",
            "value",
        ]

        return cls(
            importances=pd.concat(all_importances, axis="index", ignore_index=True)[
                ordered_columns
            ],
            report_type=report_type,
        )

    @staticmethod
    def _get_columns_to_groupby(*, frame: pd.DataFrame) -> list[str]:
        """Get the available columns from which to group by."""
        columns_to_groupby = list[str]()
        if "label" in frame.columns and frame["label"].nunique() > 1:
            columns_to_groupby.append("label")
        if "output" in frame.columns and frame["output"].nunique() > 1:
            columns_to_groupby.append("output")
        if "split" in frame.columns and frame["split"].nunique() > 1:
            columns_to_groupby.append("split")
        return columns_to_groupby

    @DisplayMixin.style_plot
    def plot(
        self,
        *,
        metric: str | None = None,
        subplot_by: str | None = "auto",
    ) -> None:
        """Plot the permutation importance.

        Parameters
        ----------
        metric : str or None, default=None
            Metric to plot. Required when multiple metrics are computed.

        subplot_by : str or None, default="auto"
            Column to use for subplotting. The possible values are:

            - if `"auto"`, depending of the information available, a meaningful decision
              is made to create subplots.
            - if a string, the corresponding column of the dataframe is used to create
              several subplots. Those plots will be organized in a grid of a single
              row and several columns.
            - if `None`, all information is plotted on a single plot. An error is raised
              if there is too much information to plot on a single plot.

        """
        return self._plot(subplot_by=subplot_by, metric=metric)

    def _plot_matplotlib(
        self,
        *,
        metric: str | None = None,
        subplot_by: str | None = "auto",
    ) -> None:
        """Dispatch the plotting function for matplotlib backend."""
        boxplot_kwargs = self._default_boxplot_kwargs.copy()
        stripplot_kwargs = self._default_stripplot_kwargs.copy()

        if metric is None and self.importances["metric"].nunique() > 1:
            raise ValueError(
                "Multiple metrics cannot be plotted at once. Please select a metric"
                " to plot the associated importances using the `metric` parameter."
            )

        self._plot_single_estimator(
            subplot_by=subplot_by,
            frame=self.frame(metric=metric, aggregate=None),
            estimator_name=self.importances["estimator"].unique()[0],
            boxplot_kwargs=boxplot_kwargs,
            stripplot_kwargs=stripplot_kwargs,
        )

    @staticmethod
    def _aggregate_over_split(*, frame: pd.DataFrame) -> pd.DataFrame:
        """Compute the averaged scores over the splits."""
        group_by = frame.columns.difference(["repetition", "value"]).tolist()
        return (
            frame.drop(columns=["repetition"])
            .groupby(group_by, sort=False, dropna=False)
            .aggregate("mean")
            .reset_index()
        )

    def _plot_single_estimator(
        self,
        *,
        subplot_by: str | None,
        frame: pd.DataFrame,
        estimator_name: str,
        boxplot_kwargs: dict[str, Any],
        stripplot_kwargs: dict[str, Any],
    ) -> None:
        """Plot the permutation importance for an `EstimatorReport`."""
        aggregate_title = ""
        columns_to_groupby = self._get_columns_to_groupby(frame=frame)
        if subplot_by == "auto" or subplot_by is None:
            if "split" in columns_to_groupby:
                frame = self._aggregate_over_split(frame=frame)
                columns_to_groupby.remove("split")
                aggregate_title = " averaged over splits"

            if subplot_by == "auto":
                col = columns_to_groupby[0] if columns_to_groupby else None
                hue = None
            else:  # subplot_by is None
                col = None
                hue = columns_to_groupby[0] if columns_to_groupby else None

        else:
            if subplot_by not in columns_to_groupby:
                raise ValueError(
                    f"The column {subplot_by!r} is not available for subplotting. "
                    "You can use the following values to create subplots: "
                    f"{', '.join(columns_to_groupby + ['auto', 'None'])}"
                )

            remaining = set(columns_to_groupby) - {subplot_by}
            if "split" in remaining:
                frame = self._aggregate_over_split(frame=frame)
                remaining.remove("split")
                aggregate_title = " averaged over splits"

            col = subplot_by
            hue = next(iter(remaining), None)

        if hue is None:
            # we don't need the palette and we are at risk of raising an error or
            # deprecation warning if passing palette without a hue
            stripplot_kwargs.pop("palette", None)
        else:
            boxplot_kwargs.setdefault("palette", "tab10")

        self.facet_ = sns.catplot(
            data=frame,
            x="value",
            y="feature",
            hue=hue,
            col=col,
            kind="strip",
            dodge=True,
            sharex=False,
            **stripplot_kwargs,
        ).map_dataframe(
            sns.boxplot,
            x="value",
            y="feature",
            hue=hue,
            dodge=True,
            **boxplot_kwargs,
        )
        add_background_features = hue is not None

        metric_name = frame["metric"].unique()[0]
        self.figure_, self.ax_ = self.facet_.figure, self.facet_.axes.squeeze()
        for row_axes in self.facet_.axes:
            for ax in row_axes:
                _decorate_matplotlib_axis(
                    ax=ax,
                    add_background_features=add_background_features,
                    n_features=frame["feature"].nunique(),
                    xlabel=f"Decrease in {metric_name}",
                    ylabel="",
                )
        if len(self.ax_.flatten()) == 1:
            self.ax_ = self.ax_.flatten()[0]
        data_source = frame["data_source"].unique()[0]
        self.figure_.suptitle(
            f"Permutation importance{aggregate_title}\n"
            f"of {estimator_name} on {data_source} set"
        )

    def frame(
        self,
        *,
        metric: str | list[str] | None = None,
        aggregate: Aggregate | None = ("mean", "std"),
    ) -> pd.DataFrame:
        """Get the feature importance in a dataframe format.

        Parameters
        ----------
        metric : str or list of str, default=None
            Filter the importances by metric. If `None`, all importances associated with
            each metric are returned.

        aggregate : {"mean", "std"}, ("mean", std) or None, default=("mean", "std")
            Aggregate the importances by the given metric.

        Returns
        -------
        pd.DataFrame
            Dataframe containing the importances.
        """
        if self.report_type == "estimator":
            columns_to_drop = ["estimator", "split"]
            group_by = ["data_source", "metric", "feature"]
        elif self.report_type == "cross-validation":
            columns_to_drop = ["estimator"]
            group_by = ["data_source", "metric", "split", "feature"]
        else:
            raise TypeError(f"Unexpected report type: {self.report_type!r}")

        frame = self.importances.copy()
        if metric is not None:
            frame = frame.query("metric in @metric")

        if frame["label"].isna().all():
            # regression problem or averaged classification metric
            columns_to_drop.append("label")
        else:
            group_by.append("label")
        if frame["output"].isna().all():
            # classification problem or averaged regression metric
            columns_to_drop.append("output")
        else:
            group_by.append("output")

        frame = frame.drop(columns=columns_to_drop)

        if aggregate is not None:
            frame = (
                frame.drop(columns=["repetition"])
                # avoid sorting the features by name and do not drop NA from
                # output or labels in case of mixed metrics (i.e. averaged vs.
                # non-averaged)
                .groupby(group_by, sort=False, dropna=False)
                .aggregate(aggregate)
            ).reset_index()
            if isinstance(frame.columns, pd.MultiIndex):
                frame.columns = flatten_multi_index(frame.columns)
        return frame

    # ignore the type signature because we override kwargs by specifying the name of
    # the parameters for the user.
    def set_style(  # type: ignore[override]
        self,
        *,
        policy: Literal["override", "update"] = "update",
        boxplot_kwargs: dict[str, Any] | None = None,
        stripplot_kwargs: dict[str, Any] | None = None,
    ):
        """Set the style parameters for the display.

        Parameters
        ----------
        policy : {"override", "update"}, default="update"
            Policy to use when setting the style parameters.
            If "override", existing settings are set to the provided values.
            If "update", existing settings are not changed; only settings that were
            previously unset are changed.

        boxplot_kwargs : dict, default=None
            Keyword arguments to be passed to :func:`seaborn.boxplot` for
            rendering the importances with a :class:`~skore.EstimatorReport`.

        stripplot_kwargs : dict, default=None
            Keyword arguments to be passed to :func:`seaborn.stripplot` for
            rendering the importances with a :class:`~skore.EstimatorReport`.

        Returns
        -------
        self : object
            Returns the instance itself.

        Raises
        ------
        ValueError
            If a style parameter is unknown.
        """
        return super().set_style(
            policy=policy,
            boxplot_kwargs=boxplot_kwargs or {},
            stripplot_kwargs=stripplot_kwargs or {},
        )



from __future__ import annotations

from typing import TYPE_CHECKING, cast

import numpy as np
from sklearn.utils.metaestimators import available_if

from skore._externals._pandas_accessors import DirNamesMixin
from skore._sklearn._base import _BaseAccessor
from skore._sklearn._plot.inspection.coefficients import CoefficientsDisplay
from skore._sklearn._plot.inspection.impurity_decrease import ImpurityDecreaseDisplay
from skore._utils._accessor import (
    _check_comparison_report_sub_estimators_have_coef,
    _check_comparison_report_sub_estimators_have_feature_importances,
)

if TYPE_CHECKING:
    from skore import ComparisonReport
    from skore._sklearn._cross_validation.report import CrossValidationReport


class _InspectionAccessor(_BaseAccessor["ComparisonReport"], DirNamesMixin):
    """Accessor for model inspection related operations.

    You can access this accessor using the `inspection` attribute.
    """

    def __init__(self, parent: ComparisonReport) -> None:
        super().__init__(parent)

    @available_if(_check_comparison_report_sub_estimators_have_coef())
    def coefficients(self) -> CoefficientsDisplay:
        """Retrieve the coefficients for each report, including the intercepts.

        Comparison reports with the same features are put under one key and are plotted
        together. When some reports share the same features and others do not, those
        with the same features are plotted together.

        Returns
        -------
        :class:`CoefficientsDisplay`
            The feature importance display containing model coefficients and
            intercept.

        Examples
        --------
        >>> from sklearn.datasets import load_diabetes
        >>> from sklearn.linear_model import Ridge
        >>> from skore import train_test_split
        >>> from skore import ComparisonReport, EstimatorReport
        >>> X, y = load_diabetes(return_X_y=True)
        >>> split_data = train_test_split(X=X, y=y, shuffle=False, as_dict=True)
        >>> report_big_alpha = EstimatorReport(Ridge(alpha=1e3), **split_data)
        >>> report_small_alpha = EstimatorReport(Ridge(alpha=1e-3), **split_data)
        >>> report = ComparisonReport({
        ...     "report small alpha": report_small_alpha,
        ...     "report big alpha": report_big_alpha,
        ... })
        >>> display = report.inspection.coefficients()
        >>> display.frame()
                     estimator     feature  coefficient
        0   report small alpha   Intercept      151.5...
        1   report small alpha  Feature #0      -11.6...
        2   report small alpha  Feature #1     -238.2...
        3   report small alpha  Feature #2      505.5...
        4   report small alpha  Feature #3      298.4...
        5   report small alpha  Feature #4     -408.5...
        6   report small alpha  Feature #5      164.0...
        7   report small alpha  Feature #6      -55.1...
        8   report small alpha  Feature #7      123.0...
        9   report small alpha  Feature #8      633.8...
        10  report small alpha  Feature #9       99.4...
        11    report big alpha   Intercept      151.0...
        12    report big alpha  Feature #0        0.2...
        13    report big alpha  Feature #1        0.0...
        14    report big alpha  Feature #2        0.6...
        15    report big alpha  Feature #3        0.5...
        16    report big alpha  Feature #4        0.2...
        17    report big alpha  Feature #5        0.2...
        18    report big alpha  Feature #6       -0.4...
        19    report big alpha  Feature #7        0.5...
        20    report big alpha  Feature #8        0.6...
        21    report big alpha  Feature #9        0.4...
        >>> display.plot() # shows plot
        """
        if self._parent._report_type == "comparison-estimator":
            return CoefficientsDisplay._compute_data_for_display(
                estimators=[
                    report.estimator_ for report in self._parent.reports_.values()
                ],
                names=list(self._parent.reports_.keys()),
                splits=[np.nan] * len(self._parent.reports_),
                report_type=self._parent._report_type,
            )
        else:  # self._parent._report_type == "comparison-cross-validation"
            estimators, names = [], []
            splits: list[int | float] = []
            for name, report in self._parent.reports_.items():
                cross_validation_report = cast("CrossValidationReport", report)
                for split_idx, estimator_report in enumerate(
                    cross_validation_report.estimator_reports_
                ):
                    estimators.append(estimator_report.estimator_)
                    names.append(name)
                    splits.append(split_idx)
            return CoefficientsDisplay._compute_data_for_display(
                estimators=estimators,
                names=names,
                splits=splits,
                report_type=self._parent._report_type,
            )

    @available_if(_check_comparison_report_sub_estimators_have_feature_importances())
    def impurity_decrease(self) -> ImpurityDecreaseDisplay:
        """Retrieve the Mean Decrease in Impurity (MDI) for each report.

        This method is available for estimators that expose a `feature_importances_`
        attribute. See for example
        :attr:`sklearn.ensemble.GradientBoostingClassifier.inspections_`.

        In particular, note that the MDI is computed at fit time, i.e. using the
        training data.

        Comparison reports with the same features are put under one key and are plotted
        together. When some reports share the same features and others do not, those
        with the same features are plotted together.

        Returns
        -------
        :class:`ImpurityDecreaseDisplay`
            The impurity decrease display containing the feature importances.

        Examples
        --------
        >>> from sklearn.datasets import load_iris
        >>> from sklearn.ensemble import RandomForestClassifier
        >>> from skore import train_test_split
        >>> from skore import ComparisonReport, EstimatorReport
        >>> X, y = load_iris(return_X_y=True, as_frame=True)
        >>> split_data = train_test_split(X=X, y=y, shuffle=False, as_dict=True)
        >>> report_small_trees = EstimatorReport(
        ...     RandomForestClassifier(max_depth=2, random_state=0), **split_data
        ... )
        >>> report_big_trees = EstimatorReport(
        ...     RandomForestClassifier(random_state=0), **split_data
        ... )
        >>> report = ComparisonReport({
        ...     "small trees": report_small_trees,
        ...     "big trees": report_big_trees,
        ... })
        >>> display = report.inspection.impurity_decrease()
        >>> display.frame()
             estimator            feature   importance
        0  small trees  sepal length (cm)       0.1...
        1  small trees   sepal width (cm)      0.0...
        2  small trees  petal length (cm)       0.4...
        3  small trees   petal width (cm)       0.4...
        4    big trees  sepal length (cm)       0.0...
        5    big trees   sepal width (cm)       0.0...
        6    big trees  petal length (cm)       0.4...
        7    big trees   petal width (cm)       0.4...
        >>> display.plot() # shows plot
        """
        if self._parent._report_type == "comparison-estimator":
            return ImpurityDecreaseDisplay._compute_data_for_display(
                estimators=[
                    report.estimator_ for report in self._parent.reports_.values()
                ],
                names=list(self._parent.reports_.keys()),
                splits=[np.nan] * len(self._parent.reports_),
                report_type=self._parent._report_type,
            )
        else:  # self._parent._report_type == "comparison-cross-validation":
            estimators, names = [], []
            splits: list[int | float] = []
            for name, report in self._parent.reports_.items():
                report = cast("CrossValidationReport", report)
                for split_idx, estimator_report in enumerate(report.estimator_reports_):
                    estimators.append(estimator_report.estimator_)
                    names.append(name)
                    splits.append(split_idx)
            return ImpurityDecreaseDisplay._compute_data_for_display(
                estimators=estimators,
                names=names,
                splits=splits,
                report_type=self._parent._report_type,
            )

    ####################################################################################
    # Methods related to the help tree
    ####################################################################################

    def __repr__(self) -> str:
        """Return a string representation using rich."""
        return self._rich_repr(class_name="skore.EstimatorReport.inspection")



from __future__ import annotations

from collections.abc import Callable

from numpy.typing import ArrayLike
from sklearn.utils.metaestimators import available_if

from skore._externals._pandas_accessors import DirNamesMixin
from skore._sklearn._base import _BaseAccessor
from skore._sklearn._cross_validation.report import CrossValidationReport
from skore._sklearn._plot.inspection.coefficients import CoefficientsDisplay
from skore._sklearn._plot.inspection.impurity_decrease import ImpurityDecreaseDisplay
from skore._sklearn._plot.inspection.permutation_importance import (
    PermutationImportanceDisplay,
)
from skore._sklearn.types import DataSource
from skore._utils._accessor import (
    _check_cross_validation_sub_estimator_has_coef,
    _check_cross_validation_sub_estimator_has_feature_importances,
)
from skore._utils._cache_key import deep_key_sanitize

Metric = str | Callable | list[str] | tuple[str] | dict[str, Callable] | None


class _InspectionAccessor(_BaseAccessor[CrossValidationReport], DirNamesMixin):
    """Accessor for model inspection related operations.

    You can access this accessor using the `inspection` attribute.
    """

    def __init__(self, parent: CrossValidationReport) -> None:
        super().__init__(parent)

    @available_if(_check_cross_validation_sub_estimator_has_coef())
    def coefficients(self) -> CoefficientsDisplay:
        """Retrieve the coefficients across splits, including the intercept.

        Returns
        -------
        :class:`CoefficientsDisplay`
            The feature importance display containing model coefficients and
            intercept.

        Examples
        --------
        >>> from sklearn.datasets import make_regression
        >>> from sklearn.linear_model import Ridge
        >>> from skore import CrossValidationReport
        >>> X, y = make_regression(n_features=3, random_state=42)
        >>> report = CrossValidationReport(estimator=Ridge(), X=X, y=y, splitter=2)
        >>> display = report.inspection.coefficients()
        >>> display.frame()
            split     feature   coefficient
        0       0   Intercept       -0.1...
        1       0  Feature #0       73.2...
        2       0  Feature #1       26.6...
        3       0  Feature #2       17.1...
        4       1   Intercept        0.2...
        5       1  Feature #0       73.8...
        6       1  Feature #1       27.4...
        7       1  Feature #2       17.1...
        >>> display.plot() # shows plot
        """
        return CoefficientsDisplay._compute_data_for_display(
            estimators=[
                report.estimator_ for report in self._parent.estimator_reports_
            ],
            names=[
                report.estimator_name_ for report in self._parent.estimator_reports_
            ],
            splits=list(range(len(self._parent.estimator_reports_))),
            report_type=self._parent._report_type,
        )

    def permutation_importance(
        self,
        *,
        data_source: DataSource = "test",
        X: ArrayLike | None = None,
        y: ArrayLike | None = None,
        at_step: int | str = 0,
        metric: Metric = None,
        n_repeats: int = 5,
        max_samples: float = 1.0,
        n_jobs: int | None = None,
        seed: int | None = None,
    ) -> PermutationImportanceDisplay:
        """Display the permutation feature importance.

        This computes the permutation importance using sklearn's
        :func:`~sklearn.inspection.permutation_importance` function,
        which consists in permuting the values of one feature and comparing
        the value of `metric` between with and without the permutation, which gives an
        indication on the impact of the feature.

        By default, `seed` is set to `None`, which means the function will
        return a different result at every call. In that case, the results are not
        cached. If you wish to take advantage of skore's caching capabilities, make
        sure you set the `seed` parameter.

        Parameters
        ----------
        data_source : {"test", "train", "X_y"}, default="test"
            The data source to use.

            - "test" : use the test set provided when creating the report.
            - "train" : use the train set provided when creating the report.
            - "X_y" : use the provided `X` and `y` to compute the metric.

        X : array-like of shape (n_samples, n_features), default=None
            New data on which to compute the metric. By default, we use the test
            set provided when creating the report.

        y : array-like of shape (n_samples,), default=None
            New target on which to compute the metric. By default, we use the test
            target provided when creating the report.

        at_step : int or str, default=0
            If the estimator is a :class:`~sklearn.pipeline.Pipeline`, at which step of
            the pipeline the importance is computed. If `n`, then the features that
            are evaluated are the ones *right before* the `n`-th step of the pipeline.
            For instance,

            - If 0, compute the importance just before the start of the pipeline (i.e.
              the importance of the raw input features).
            - If -1, compute the importance just before the end of the pipeline (i.e.
              the importance of the fully engineered features, just before the actual
              prediction step).

            If a string, will be searched among the pipeline's `named_steps`.

            Has no effect if the estimator is not a :class:`~sklearn.pipeline.Pipeline`.

        metric : str, callable, list, tuple, or dict, default=None
            The metric to pass to :func:`~sklearn.inspection.permutation_importance`.

            If `metric` represents a single metric, one can use:

            - a single string, which must be one of the supported metrics;
            - a callable that returns a single value.

            If `metric` represents multiple metrics, one can use:

            - a list or tuple of unique strings, which must be one of the supported
              metrics;
            - a callable returning a dictionary where the keys are the metric names
              and the values are the metric scores;
            - a dictionary with metric names as keys and callables a values.

        n_repeats : int, default=5
            Number of times to permute a feature.

        max_samples : int or float, default=1.0
            The number of samples to draw from `X` to compute feature importance
            in each repeat (without replacement).

            - If int, then draw max_samples samples.
            - If float, then draw max_samples * X.shape[0] samples.
            - If max_samples is equal to 1.0 or X.shape[0], all samples will be used.

            While using this option may provide less accurate importance estimates,
            it keeps the method tractable when evaluating feature importance on
            large datasets. In combination with n_repeats, this allows to control
            the computational speed vs statistical accuracy trade-off of this method.

        n_jobs : int or None, default=None
            Number of jobs to run in parallel. -1 means using all processors.

        seed : int or None, default=None
            The seed used to initialize the random number generator used for the
            permutations.

        Returns
        -------
        :class:`PermutationImportanceDisplay`
            The permutation importance display.

        Examples
        --------
        >>> from sklearn.datasets import make_regression
        >>> from sklearn.linear_model import Ridge
        >>> from skore import train_test_split
        >>> from skore import CrossValidationReport
        >>> X, y = make_regression(n_features=3, random_state=0)
        >>> report = CrossValidationReport(estimator=Ridge(), X=X, y=y, splitter=2)
        >>> report.inspection.permutation_importance(
        ...    n_repeats=2,
        ...    seed=0,
        ... ).frame(aggregate=None)
           data_source metric  split     feature  repetition     value
        0         test     r2      0  Feature #0           1  0.71...
        1         test     r2      0  Feature #1           1  1.59...
        2         test     r2      0  Feature #2           1  0.01...
        3         test     r2      0  Feature #0           2  0.70...
        4         test     r2      0  Feature #1           2  1.58...
        5         test     r2      0  Feature #2           2  0.01...
        6         test     r2      1  Feature #0           1  0.63...
        7         test     r2      1  Feature #1           1  1.82...
        8         test     r2      1  Feature #2           1  0.01...
        9         test     r2      1  Feature #0           2  0.49...
        10        test     r2      1  Feature #1           2  1.15...
        11        test     r2      1  Feature #2           2  0.01...
        >>> report.inspection.permutation_importance(
        ...    metric=["r2", "neg_mean_squared_error"],
        ...    n_repeats=2,
        ...    seed=0,
        ... ).frame(aggregate=None)
           data_source                  metric  ...  repetition         value
        0         test                      r2  ...           1      0.71...
        1         test                      r2  ...           1      1.59...
        2         test                      r2  ...           1      0.01...
        3         test                      r2  ...           2      0.70...
        4         test                      r2  ...           2      1.58...
        5         test                      r2  ...           2      0.01...
        ...
        >>> report.inspection.permutation_importance(
        ...    n_repeats=2,
        ...    seed=0,
        ... ).frame()
          data_source metric  split     feature  value_mean  value_std
        0        test     r2      0  Feature #0     0.71...    0.00...
        1        test     r2      0  Feature #1     1.58...    0.00...
        2        test     r2      0  Feature #2     0.01...    0.00...
        3        test     r2      1  Feature #0     0.56...    0.09...
        4        test     r2      1  Feature #1     1.49...    0.47...
        5        test     r2      1  Feature #2     0.01...    0.00...
        >>> # Compute the importance at the end of feature engineering pipeline
        >>> from sklearn.pipeline import make_pipeline
        >>> from sklearn.preprocessing import StandardScaler
        >>> pipeline = make_pipeline(StandardScaler(), Ridge())
        >>> pipeline_report = CrossValidationReport(
        ...     estimator=pipeline, X=X, y=y, splitter=2
        ... )
        >>> pipeline_report.inspection.permutation_importance(
        ...    n_repeats=2,
        ...    seed=0,
        ...    at_step=-1,
        ... ).frame()
          data_source metric  split feature  value_mean  value_std
        0        test     r2      0      x0    0.70...     0.00...
        1        test     r2      0      x1    1.58...     0.00...
        2        test     r2      0      x2    0.01...     0.00...
        3        test     r2      1      x0    0.56...     0.09...
        4        test     r2      1      x1    1.49...     0.47...
        5        test     r2      1      x2    0.01...     0.00...
        >>> pipeline_report.inspection.permutation_importance(
        ...    n_repeats=2,
        ...    seed=0,
        ...    at_step="ridge",
        ... ).frame()
          data_source metric  split feature  value_mean  value_std
        0        test     r2      0      x0    0.70...     0.00...
        1        test     r2      0      x1    1.58...     0.00...
        2        test     r2      0      x2    0.01...     0.00...
        3        test     r2      1      x0    0.56...     0.09...
        4        test     r2      1      x1    1.49...     0.47...
        5        test     r2      1      x2    0.01...     0.00...

        Notes
        -----
        Even if pipeline components output sparse arrays, these will be made dense.
        """  # noqa: E501
        if data_source == "X_y":
            X_, y_true, data_source_hash = self._get_X_y_and_data_source_hash(
                data_source=data_source, X=X, y=y
            )
        else:
            data_source_hash = None

        # NOTE: to temporary improve the `project.put` UX, we always store the
        # permutation importance into the cache dictionary even when seed is None.
        # Be aware that if seed is None, we still trigger the computation for all cases.
        # We only store it such that when we serialize to send to the hub, we only
        # fetch for the cache store instead of recomputing it because it is expensive.
        # FIXME: the workaround above should be removed once we are able to trigger
        # computation on the server side of skore-hub.

        if seed is not None and not isinstance(seed, int):
            raise ValueError(f"seed must be an integer or None; got {type(seed)}")

        # n_jobs should not be in cache
        kwargs = {"n_repeats": n_repeats, "max_samples": max_samples, "seed": seed}
        cache_key = deep_key_sanitize(
            (
                self._parent._hash,
                "permutation_importance",
                data_source,
                at_step,
                data_source_hash,
                metric,
                kwargs,
            )
        )

        # NOTE: avoid to fetch from the cache if the seed is None because we want
        # to trigger the computation in this case. We only have the permutation
        # stored as a workaround for the serialization for skore-hub as explained
        # earlier.
        display = None if seed is None else self._parent._cache.get(cache_key)
        if display is None:
            Xs: list[ArrayLike] = []
            ys: list[ArrayLike] = []
            for report in self._parent.estimator_reports_:
                if data_source == "X_y":
                    Xs.append(X_)
                    ys.append(y_true)
                else:
                    report_X, report_y, _ = (
                        report.inspection._get_X_y_and_data_source_hash(
                            data_source=data_source
                        )
                    )
                    Xs.append(report_X)
                    ys.append(report_y)

            display = PermutationImportanceDisplay._compute_data_for_display(
                data_source=data_source,
                estimators=[
                    report.estimator_ for report in self._parent.estimator_reports_
                ],
                names=[
                    report.estimator_name_ for report in self._parent.estimator_reports_
                ],
                splits=list(range(len(self._parent.estimator_reports_))),
                Xs=Xs,
                ys=ys,
                at_step=at_step,
                metric=metric,
                n_repeats=n_repeats,
                max_samples=max_samples,
                n_jobs=n_jobs,
                seed=seed,
                report_type="cross-validation",
            )

            if cache_key is not None:
                # NOTE: for the moment, we will always store the permutation importance
                self._parent._cache[cache_key] = display

        return display

    @available_if(_check_cross_validation_sub_estimator_has_feature_importances())
    def impurity_decrease(self) -> ImpurityDecreaseDisplay:
        """Retrieve the Mean Decrease in Impurity (MDI) for each split.

        This method is available for estimators that expose a `feature_importances_`
        attribute. See for example
        :attr:`sklearn.ensemble.GradientBoostingClassifier.feature_importances_`.

        In particular, note that the MDI is computed at fit time, i.e. using the
        training data.

        Returns
        -------
        :class:`ImpurityDecreaseDisplay`
            The impurity decrease display containing the feature importances.

        Examples
        --------
        >>> from sklearn.datasets import load_iris
        >>> from sklearn.ensemble import RandomForestClassifier
        >>> from skore import CrossValidationReport
        >>> iris = load_iris(as_frame=True)
        >>> X, y = iris.data, iris.target
        >>> y = iris.target_names[y]
        >>> report = CrossValidationReport(
        ...     estimator=RandomForestClassifier(random_state=0), X=X, y=y, splitter=5
        ... )
        >>> display = report.inspection.impurity_decrease()
        >>> display.frame()
            split            feature  importance
        0       0  sepal length (cm)       0.0...
        1       0   sepal width (cm)       0.0...
        2       0  petal length (cm)       0.4...
        3       0   petal width (cm)       0.4...
        4       1  sepal length (cm)       0.0...
        ...
        >>> display.plot() # shows plot
        """
        return ImpurityDecreaseDisplay._compute_data_for_display(
            estimators=[
                report.estimator_ for report in self._parent.estimator_reports_
            ],
            names=[
                report.estimator_name_ for report in self._parent.estimator_reports_
            ],
            splits=list(range(len(self._parent.estimator_reports_))),
            report_type=self._parent._report_type,
        )

    ####################################################################################
    # Methods related to the help tree
    ####################################################################################

    def __repr__(self) -> str:
        """Return a string representation using rich."""
        return self._rich_repr(class_name="skore.CrossValidationReport.inspection")
